# -*- coding: utf-8 -*-
from __future__ import absolute_import

import json
import time
import inspect
import os

import quartzbio

from .version import VERSION
from .errors import QuartzBioError
from .utils.validators import validate_api_host_url

import platform
import requests
import textwrap
import logging

from requests import Session
from requests import codes
from requests.auth import AuthBase
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from six.moves.urllib.parse import urljoin

# Try using pyopenssl if available.
# Requires: pip install pyopenssl ndg-httpsclient pyasn1
# See http://urllib3.readthedocs.org/en/latest/contrib.html#module-urllib3.contrib.pyopenssl  # noqa
try:
    import urllib3.contrib.pyopenssl

    urllib3.contrib.pyopenssl.inject_into_urllib3()
except ImportError:
    pass

logger = logging.getLogger("quartzbio")


def _handle_api_error(response):
    if response.status_code not in [400, 401, 403, 404]:
        logger.info("API Error: %d" % response.status_code)
    raise QuartzBioError(response=response)


def _handle_request_error(e):
    if isinstance(e, requests.exceptions.RequestException):
        msg = QuartzBioError.default_message
        err = "%s: %s" % (type(e).__name__, str(e))
    else:
        msg = (
            "Unexpected error communicating with QuartzBio.\n"
            "It looks like there's probably a configuration "
            "issue locally.\nIf this problem persists, let us "
            "know at support@quartzbio.com."
        )
        err = "A %s was raised" % (type(e).__name__,)
        if str(e):
            err += " with error message %s" % (str(e),)
        else:
            err += " with no error message"
    msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
    raise QuartzBioError(message=msg)


class QuartzBioTokenAuth(AuthBase):
    """Custom auth handler for QuartzBio API token authentication"""

    def __init__(self, token=None, token_type="Token"):
        self.token = token
        self.token_type = token_type

        if not self.token:
            # Prefer the OAuth2 access token over the API key.
            if quartzbio.access_token:
                self.token_type = "Bearer"
                self.token = quartzbio.access_token
            elif quartzbio.api_key:
                self.token_type = "Token"
                self.token = quartzbio.api_key

    def __call__(self, r):
        if self.token:
            r.headers["Authorization"] = "{0} {1}".format(self.token_type, self.token)
        return r

    def __repr__(self):
        if self.token:
            return self.token_type
        else:
            return "Anonymous"


class QuartzBioClient(object):
    """A requests-based HTTP client for QuartzBio API resources"""

    def __init__(
        self,
        host=None,
        token=None,
        token_type="Token",
        include_resources=True,
        retry_all=None,
    ):
        self.set_host(host)
        self.set_token(token, token_type)
        self._headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Accept-Encoding": "gzip,deflate",
        }
        self.set_user_agent()

        if retry_all is None:
            retry_all = bool(os.environ.get("QUARTZBIO_RETRY_ALL"))

        if bool(retry_all):
            logger.info("Retries enabled for all API requests")
            allowed_methods = frozenset(
                ["HEAD", "GET", "PUT", "POST", "PATCH", "DELETE", "OPTIONS", "TRACE"]
            )

            retries = Retry(
                total=5,
                backoff_factor=2,
                status_forcelist=[
                    codes.bad_gateway,
                    codes.service_unavailable,
                    codes.gateway_timeout,
                ],
                allowed_methods=allowed_methods,
            )
        else:
            logger.info("Retries enabled for read-only API requests")
            retries = Retry(
                total=5,
                backoff_factor=2,
                status_forcelist=[
                    codes.bad_gateway,
                    codes.service_unavailable,
                    codes.gateway_timeout,
                ],
            )

        # Use a session with a retry policy to handle
        # intermittent connection errors.
        adapter = HTTPAdapter(max_retries=retries)
        self._session = Session()
        self._session.mount(self._host, adapter)

        # Import all resources into the client
        if include_resources:
            skip = (
                "QuartzBioError",
                "QuartzBioClient",
            )
            for name, class_ in inspect.getmembers(quartzbio, inspect.isclass):
                if name in skip:
                    continue
                subclass = type(name, (class_,), {"_client": self})
                setattr(self, name, subclass)

    def set_host(self, host=None):
        self._host = validate_api_host_url(host or quartzbio.api_host)
        # If the domain ends with .quartzbio.com, determine if
        # we are being redirected. If so, update the url with the new host
        # and log a warning.
        if self._host and self._host.rstrip("/").endswith(".api.quartzbio.com"):
            old_host = self._host.rstrip("/")
            response = requests.head(old_host, allow_redirects=True)
            # Strip the port number from the host for comparison
            new_host = (
                validate_api_host_url(response.url).rstrip("/").replace(":443", "")
            )
            if old_host != new_host:
                logger.warn(
                    'API host redirected from "{}" to "{}", '
                    "please update your local credentials file".format(
                        old_host, new_host
                    )
                )
                self._host = new_host

    def set_token(self, token=None, token_type="Token"):
        self._auth = QuartzBioTokenAuth(token, token_type)

    def set_user_agent(self, name=None, version=None):
        ua = "quartzbio-python-client/{} python-requests/{} {}/{}".format(
            VERSION,
            requests.__version__,
            platform.python_implementation(),
            platform.python_version(),
        )

        # Prefix the name of the app or script before the
        # default user-agent.
        if name:
            name = name.replace(" ", "-")
            if version:
                ua = "{}/{} {}".format(name, version, ua)
            else:
                ua = "{} {}".format(name, ua)

        self._headers["User-Agent"] = ua

    def whoami(self):
        return self.get("/v1/user", {})

    def get(self, url, params, **kwargs):
        """Issues an HTTP GET across the wire via the Python requests
        library. See *request()* for information on keyword args."""
        kwargs["params"] = params
        return self.request("GET", url, **kwargs)

    def post(self, url, data, **kwargs):
        """Issues an HTTP POST across the wire via the Python requests
        library. See *request* for information on keyword args."""
        kwargs["data"] = data
        return self.request("POST", url, **kwargs)

    def delete(self, url, data, **kwargs):
        """Issues an HTTP DELETE across the wire via the Python requests
        library. See *request* for information on keyword args."""
        kwargs["data"] = data
        return self.request("DELETE", url, **kwargs)

    def request(self, method, url, **kwargs):
        """
        Issues an HTTP Request across the wire via the Python requests
        library.

        Parameters
        ----------

        method : str
           an HTTP method: GET, PUT, POST, DELETE, ...

        url : str
           the place to connect to. If the url doesn't start
           with a protocol (https:// or http://), we'll slap
                quartzbio.api_host in the front.

        allow_redirects: bool, optional
           set *False* we won't follow any redirects

        headers: dict, optional

          Custom headers can be provided here; generally though this
          will be set correctly by default dependent on the
          method type. If the content type is JSON, we'll
          JSON-encode params.

        param : dict, optional
           passed as *params* in the requests.request

        timeout : int, optional
          timeout value in seconds for the request

        raw: bool, optional
          unless *True* the response encoded to json

        files: file
          File content in the form of a file handle which is to be
          uploaded. Files are passed in POST requests

        Returns
        -------
        response object. If *raw* is not *True* and
        repsonse if valid the object will be JSON encoded. Otherwise
        it will be the request.reposne object.
        """

        opts = {
            "allow_redirects": True,
            "auth": self._auth,
            "data": {},
            "files": None,
            "headers": dict(self._headers),
            "params": {},
            "timeout": 80,
            "verify": True,
        }

        raw = kwargs.pop("raw", False)
        debug = kwargs.pop("debug", False)
        opts.update(kwargs)
        method = method.upper()

        if opts["files"]:
            # Don't use application/json for file uploads or GET requests
            opts["headers"].pop("Content-Type", None)
        else:
            opts["data"] = json.dumps(opts["data"])

        if not url.startswith(self._host):
            url = urljoin(self._host, url)

        logger.debug("API %s Request: %s" % (method, url))

        if debug:
            self._log_raw_request(method, url, **opts)

        try:
            response = self._session.request(method, url, **opts)
        except Exception as e:
            _handle_request_error(e)

        if 429 == response.status_code:
            delay = int(response.headers["retry-after"]) + 1
            logger.warn("Too many requests. Retrying in {0}s.".format(delay))
            time.sleep(delay)
            return self.request(method, url, **kwargs)

        if not (200 <= response.status_code < 400):
            _handle_api_error(response)

        # 204 is used on deletion. There is no JSON here.
        if raw or response.status_code in [204, 301, 302]:
            return response

        try:
            return response.json()
        except Exception:
            raise QuartzBioError(
                "Could not parse JSON response: {}".format(response.content)
            )

    def _log_raw_request(self, method, url, **kwargs):
        from requests import Request, Session

        req = Request(
            method=method.upper(), url=url, data=kwargs["data"], params=kwargs["params"]
        )
        prepped = Session().prepare_request(
            req,
        )
        logger.debug(prepped.headers)
        logger.debug(prepped.body)

    def __repr__(self):
        return "<QuartzBioClient {0} {1}>".format(self._host, self._auth)


client = QuartzBioClient(include_resources=False)



# -*- coding: utf-8 -*-
"""
QuartzBio Python Client
~~~~~~~~~~~~~~~~~~~~~~

This is the Python client & library for the QuartzBio API.

Have questions or comments? email us at: support@quartzbio.com
"""

from __future__ import absolute_import
from __future__ import print_function

__docformat__ = "restructuredtext"

import os as _os
import logging as _logging

# Capture warnings (specifically from urllib3)
try:
    _logging.captureWarnings(True)
except:
    # Python 2.6 doesn't support this
    pass

# Read/Write API key
api_key = (
    _os.environ.get("SOLVEBIO_API_KEY", None)
    or _os.environ.get("QUARTZBIO_API_KEY", None)
    or _os.environ.get("EDP_API_KEY", None)
)
# OAuth2 access tokens
access_token = (
    _os.environ.get("SOLVEBIO_ACCESS_TOKEN", None)
    or _os.environ.get("QUARTZBIO_ACCESS_TOKEN", None)
    or _os.environ.get("EDP_ACCESS_TOKEN", None)
)
api_host = (
    _os.environ.get("QUARTZBIO_API_HOST", None)
    or _os.environ.get("QUARTZBIO_API_HOST", None)
    or _os.environ.get("EDP_API_HOST", None)
)


def _init_logging():
    loglevel_base = _os.environ.get("QUARTZBIO_LOGLEVEL", None) or _os.environ.get(
        "QUARTZBIO_LOGLEVEL", "WARN"
    )
    loglevel_stream = loglevel_base
    logfile = (
        _os.environ.get("QUARTZBIO_LOGFILE", None)
        or _os.environ.get("QUARTZBIO_LOGFILE")
        or "~/.quartzbio/quartzbio.log"
    )
    loglevel_file = loglevel_base

    base_logger = _logging.getLogger("quartzbio")
    base_logger.setLevel(loglevel_base)

    # clear handlers if any exist
    handlers = base_logger.handlers[:]
    for handler in handlers:
        base_logger.removeHandler(handler)
        handler.close()

    if loglevel_stream:
        stream_handler = _logging.StreamHandler()
        stream_handler.setLevel(loglevel_stream)
        stream_fmt = _logging.Formatter("[QuartzBio] %(message)s")
        stream_handler.setFormatter(stream_fmt)
        base_logger.addHandler(stream_handler)

    if logfile:
        logfile_path = _os.path.expanduser(logfile)
        logdir = _os.path.dirname(logfile_path)

        if not _os.path.isdir(logdir):
            # Handle a race condition here when running
            # multiple services that import this package.
            try:
                _os.makedirs(logdir)
            except OSError as err:
                # Re-raise anything other than 'File exists'.
                if err[1] != "File exists":
                    raise err

        file_handler = _logging.FileHandler(logfile_path)
        file_handler.setLevel(loglevel_file)
        file_fmt = _logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        file_handler.setFormatter(file_fmt)
        base_logger.addHandler(file_handler)

    try:
        base_logger.addHandler(_logging.NullHandler())
    except:
        # supports Python < 2.7
        class NullHandler(_logging.Handler):
            def emit(self, record):
                pass

        base_logger.addHandler(NullHandler())


_init_logging()

from .version import VERSION  # noqa
from .errors import QuartzBioError
from .query import Query, BatchQuery, Filter, GenomicFilter
from .global_search import GlobalSearch
from .annotate import Annotator, Expression
from .client import QuartzBioClient
from .resource import (
    Application,
    Beacon,
    BeaconSet,
    Dataset,
    DatasetCommit,
    DatasetExport,
    DatasetField,
    DatasetImport,
    DatasetMigration,
    DatasetTemplate,
    DatasetRestoreTask,
    DatasetSnapshotTask,
    Group,
    Manifest,
    Object,
    User,
    Vault,
    VaultSyncTask,
    ObjectCopyTask,
    SavedQuery,
    Task,
)


def login(**kwargs):
    """
    Sets up the auth credentials using the provided key/token,
    or checks the credentials file (if no token provided).

    Lookup order:
        1. access_token
        2. api_key
        3. local credentials

    No errors are raised if no key is found.
    """
    from .cli.auth import get_credentials

    global access_token, api_key, api_host

    # Clear any existing auth keys
    access_token, api_key = None, None
    # Update the host
    api_host = kwargs.get("api_host") or api_host

    if kwargs.get("access_token"):
        access_token = kwargs.get("access_token")
    elif kwargs.get("api_key"):
        api_key = kwargs.get("api_key")
    else:
        creds = get_credentials()
        # creds = (host, email, token_type, token)
        if creds:
            api_host = creds[0]
            if creds[2] == "Bearer":
                access_token = creds[3]
            else:
                # By default, assume it is an API key.
                api_key = creds[3]

    # Always update the client host, version and agent
    from quartzbio.client import client

    client.set_host()
    client.set_user_agent(name=kwargs.get("name"), version=kwargs.get("version"))

    if not (api_key or access_token):
        return False
    else:
        # Update the client token
        client.set_token()
        return True


__all__ = [
    "Annotator",
    "Application",
    "Expression",
    "BatchQuery",
    "Beacon",
    "BeaconSet",
    "Dataset",
    "DatasetCommit",
    "DatasetField",
    "DatasetExport",
    "DatasetImport",
    "DatasetMigration",
    "DatasetTemplate",
    "DatasetRestoreTask",
    "DatasetSnapshotTask",
    "Filter",
    "GenomicFilter",
    "GlobalSearch",
    "Group",
    "Manifest",
    "Object",
    "ObjectCopyTask",
    "Query",
    "SavedQuery",
    "QuartzBioClient",
    "QuartzBioError",
    "VaultSyncTask",
    "Task",
    "Vault",
    "User",
    "VERSION",
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import quartzbio
from ..client import client
from .credentials import get_credentials
from .credentials import save_credentials
from .credentials import delete_credentials


def login_and_save_credentials(*args, **kwargs):
    """
    Domain, email and password are used to get the user's API key.
    """
    if args and args[0].api_key:
        # Handle command-line arguments if provided.
        logged_in = quartzbio.login(
            api_key=args[0].api_key, api_host=quartzbio.api_host
        )
    elif args and args[0].access_token:
        # Handle command-line arguments if provided.
        logged_in = quartzbio.login(
            access_token=args[0].access_token, api_host=quartzbio.api_host
        )
    elif quartzbio.login(**kwargs):
        logged_in = True
    else:
        logged_in = False

    if logged_in:
        # Print information about the current user
        user = client.whoami()
        print_user(user)
        save_credentials(
            user["email"].lower(),
            client._auth.token,
            client._auth.token_type,
            quartzbio.api_host,
        )
        print("Updated local credentials file.")
    else:
        print(
            "You are not logged-in. Visit "
            "https://docs.quartzbio.com/#authentication to get started."
        )


def logout(*args):
    """
    Delete's the user's locally-stored credentials.
    """
    if get_credentials():
        delete_credentials()
        print("You have been logged out.")
    else:
        print("You are not logged-in.")


def whoami(*args, **kwargs):
    """
    Prints information about the current user.
    Assumes the user is already logged-in.
    """
    try:
        user = client.whoami()
    except Exception as e:
        print("{} (code: {})".format(e.message, e.status_code))
    else:
        print_user(user)


def print_user(user):
    """
    Prints information about the current user.
    """
    email = user["email"]
    domain = user["account"]["domain"]
    print(
        'You are logged-in to the "{}" domain as {} (server: {}).'.format(
            domain, email, quartzbio.api_host
        )
    )



from __future__ import absolute_import
import os
from pydoc import pager

TUTORIAL = os.path.abspath(os.path.dirname(__file__)) + "/tutorial.md"


def print_tutorial(args):
    pager(open(TUTORIAL, "rb").read())



from __future__ import absolute_import

from six.moves.urllib.parse import urlparse

from ..errors import QuartzBioError


def validate_api_host_url(url):
    """
    Validate QuartzBio API host url.

    Valid urls must not be empty and
    must contain either HTTP or HTTPS scheme.
    """
    if not url:
        raise QuartzBioError("No QuartzBio API host is set")

    # Default to https if no scheme is set
    if "://" not in url:
        url = "https://" + url

    parsed = urlparse(url)
    if parsed.scheme not in ["http", "https"]:
        raise QuartzBioError(
            "Invalid API host: %s. " "Missing url scheme (HTTP or HTTPS)." % url
        )
    elif not parsed.netloc:
        raise QuartzBioError("Invalid API host: %s." % url)

    return parsed.geturl()
